"""
Tests borrowed from Simon Willison's Django-OpenID project:
  * https://github.com/simonw/django-openid
"""
import pytest
from django.conf import settings
from django_comments_ink import signed


def test_sign_unsign_no_unicode():
    "sign/unsign functions should not accept unicode strings"
    with pytest.raises(TypeError):
        signed.sign("\u2021")
    with pytest.raises(TypeError):
        signed.unsign("\u2021")


def test_sign_uses_correct_key():
    "If a key is provided, sign should use it; otherwise, use SECRET_KEY"
    s = b"This is a string"

    assert signed.sign(s) == (
        s + b"." + signed.base64_hmac(s, settings.SECRET_KEY.encode("ascii"))
    )
    assert signed.sign(s, b"sekrit") == (
        s + b"." + signed.base64_hmac(s, b"sekrit")
    )


def test_sign_is_reversible():
    examples = (
        b"q;wjmbk;wkmb",
        b"3098247529087",
        b"3098247:529:087:",
        b"jkw osanteuh ,rcuh nthu aou oauh ,ud du",
        "\u2019".encode("utf8"),
    )
    for example in examples:
        assert example != signed.sign(example)
        assert example == signed.unsign(signed.sign(example))


def test_unsign_detects_tampering():
    value = b"Another string"
    signed_value = signed.sign(value)
    transforms = (
        lambda s: s.upper(),
        lambda s: s + b"a",
        lambda s: b"a" + s[1:],
        lambda s: s.replace(b".", b""),
    )
    assert value == signed.unsign(signed_value)
    for transform in transforms:
        with pytest.raises(signed.BadSignature):
            signed.unsign(transform(signed_value))


def test_encode_decode():
    objects = (
        (b"a", b"tuple"),
        b"a string",
        "a unicode string \u2019",
        {"a": "dictionary"},
    )
    for o in objects:
        assert o != signed.dumps(o)
        assert o == signed.loads(signed.dumps(o))


def test_decode_detects_tampering():
    transforms = (
        lambda s: s.upper(),
        lambda s: s + b"a",
        lambda s: b"a" + s[1:],
        lambda s: s.replace(b".", b""),
    )
    value = {"foo": "bar", "baz": 1}
    encoded = signed.dumps(value)
    assert value == signed.loads(encoded)
    for transform in transforms:
        with pytest.raises(signed.BadSignature):
            signed.loads(transform(encoded))



from django.urls import path, re_path

from .views import (
    CommentCount,
    CommentCreate,
    CommentList,
    CreateReportFlag,
    PostCommentReaction,
)

urlpatterns = [
    path("comment/", CommentCreate.as_view(), name="comments-ink-api-create"),
    re_path(
        r"^(?P<content_type>\w+[-]{1}\w+)/(?P<object_pk>[-\w]+)/$",
        CommentList.as_view(),
        name="comments-ink-api-list",
    ),
    re_path(
        r"^(?P<content_type>\w+[-]{1}\w+)/(?P<object_pk>[-\w]+)/count/$",
        CommentCount.as_view(),
        name="comments-ink-api-count",
    ),
    path(
        "react/",
        PostCommentReaction.as_view(),
        name="comments-ink-api-react",
    ),
    path("flag/", CreateReportFlag.as_view(), name="comments-ink-api-flag"),
]



from django.urls import reverse
from django_comments_ink.api.views import (
    CommentCreate,
    CreateReportFlag,
    PostCommentReaction,
)
from rest_framework.test import APIRequestFactory, force_authenticate

request_factory = APIRequestFactory()


def post_comment(data, auth_user=None):
    request = request_factory.post(reverse("comments-ink-api-create"), data)
    if auth_user:
        force_authenticate(request, user=auth_user)
    view = CommentCreate.as_view()
    return view(request)


def send_reaction(method, data, auth_user=None):
    assert method is not None
    assert method in ["get", "post", "update", "put", "delete"]
    method = getattr(request_factory, method)
    request = method(reverse("comments-ink-api-react"), data)
    if auth_user:
        force_authenticate(request, user=auth_user)
    view = PostCommentReaction.as_view()
    return view(request)


def post_flag(data, auth_user=None):
    request = request_factory.post(reverse("comments-ink-api-flag"), data)
    if auth_user:
        force_authenticate(request, user=auth_user)
    view = CreateReportFlag.as_view()
    return view(request)



from __future__ import unicode_literals

from django.conf import settings

COMMENT_MAX_LENGTH = 3000

# Extra key to salt the InkCommentForm.
COMMENTS_INK_SALT = b""

# Whether comment posts should be confirmed by email.
COMMENTS_INK_CONFIRM_EMAIL = True

# From email address.
COMMENTS_INK_FROM_EMAIL = settings.DEFAULT_FROM_EMAIL

# Contact email address.
COMMENTS_INK_CONTACT_EMAIL = settings.DEFAULT_FROM_EMAIL

# Maximum Thread Level.
COMMENTS_INK_MAX_THREAD_LEVEL = 0

# Maximum Thread Level per app.model basis.
COMMENTS_INK_MAX_THREAD_LEVEL_BY_APP_MODEL = {}

# Default order to list comments in.
COMMENTS_INK_LIST_ORDER = ("thread_id", "order")

# Form class to use.
COMMENTS_INK_FORM_CLASS = "django_comments_ink.forms.InkCommentForm"

# Model to use.
COMMENTS_INK_MODEL = "django_comments_ink.models.InkComment"

# Enum class for comment reactions.
COMMENTS_INK_COMMENT_REACTIONS_ENUM = "django_comments_ink.models.ReactionEnum"

# Enum class for object reactions.
COMMENTS_INK_OBJECT_REACTIONS_ENUM = "django_comments_ink.models.ReactionEnum"

# Send HTML emails.
COMMENTS_INK_SEND_HTML_EMAIL = True

# Whether to send emails in separate threads or use the regular method.
# Set it to False to use a third-party app like django-celery-email or
# your own celery app.
COMMENTS_INK_THREADED_EMAILS = True

# Define what commenting features a pair app_label.model can have.
COMMENTS_INK_APP_MODEL_OPTIONS = {
    "default": {
        "who_can_post": "all",  # Valid values: "users", "all",
        # Function to determine whether new comments,
        # reactions, etc. should be allowed for a given object.
        "check_input_allowed": "django_comments_ink.utils.check_input_allowed",
        # Whether to display a link to flag a comment as inappropriate.
        "comment_flagging_enabled": False,
        # Whether to allow users to submit reactions on comments.
        "comment_reactions_enabled": False,
        # Whether to allow users to submit reactions on any object
        # registered as a content type.
        "object_reactions_enabled": False,
    }
}

# When True, an approval or removal operation on a comment will
# trigger a publishing or withholding operation on its nested comments.
COMMENTS_INK_PUBLISH_OR_WITHHOLD_NESTED = True


# Define a function to return the user representation. Used by
# the web API to represent user strings in response objects.
def username(u):
    return u.username


COMMENTS_INK_API_USER_REPR = username


# Makes the "Notify me about followup comments" checkbox in the
# comment form checked (True) or unchecked (False) by default.
COMMENTS_INK_DEFAULT_FOLLOWUP = False

# How many reaction buttons can be displayed
# in a row before it breaks into another row.
COMMENTS_INK_REACTIONS_ROW_LENGTH = 4

# Display up to the given number of comments in the last page to avoid
# creating another page containing only these amount of comments.
COMMENTS_INK_MAX_LAST_PAGE_ORPHANS = 10

# Number of comments per page. When <=0 pagination is disabled.
COMMENTS_INK_ITEMS_PER_PAGE = 25

# Name of the query string parameter containing the page number.
COMMENTS_INK_PAGE_QUERY_STRING_PARAM = "cpage"

# Name of the query string parameter containing the comments to fold.
COMMENTS_INK_FOLD_QUERY_STRING_PARAM = "cfold"

# All HTML elements rendered by django-comments-ink use the 'dci' CSS selector,
# defined in 'django_comments_ink/static/django_comments_ink/css/comments.css'.
# You can alter the CSS rules applied to your comments adding your own custom
# selector to the following setting. If you wanted to modify the .body of the
# comments with, say, a different padding, line-height and font color you could
# create do so by creating the following selector in your project's CSS:
#
#   .dci.dci-custom .comment-box .body {
#       padding: 8px;
#       line-height: 1.5;
#       color: #555;
#   }
#
# And adding the following setting to your settings module:
#
#    COMMENTS_INK_CSS_CUSTOM_SELECTOR = "dci dci-custom"
#
COMMENTS_INK_CSS_CUSTOM_SELECTOR = "dci"

# Name of the directory containing the theme templates.
# By default there are 3 theme directories:
#   'avatar_in_thread', 'avatar_in_header', 'twbs_5'.
#
# Use a theme by assigning any of those values to the COMMENTS_INK_THEME_DIR
# setting. You can create your own theme templates by adding a new theme
# directory to your Django project template's directory,
# under `comments/themes/<your-theme>`.
#
COMMENTS_INK_THEME_DIR = ""

# Name of the entry in settings.CACHE to use for caching django-comments-ink.
COMMENTS_INK_CACHE_NAME = "dci"

COMMENTS_INK_CACHE_KEYS = {
    # The key 'comments_qs' holds the QuerySet by the given params.
    "comments_qs": "/comments_qs/{ctype_pk}/{object_pk}/{site_id}",
    # The key 'comments_count' stores the number of
    # comments returned by the previous QuerySet.
    "comments_count": "/comments_count/{ctype_pk}/{object_pk}/{site_id}",
    # The key 'comments_paged' stores a dictionary of k: v, where
    # keys are combinations of page number and folded comments, and
    # values are cache keys where to find the computed values that
    # correspond to the output of paginate_querysey.
    "comments_paged": "/comments_paged/{ctype_pk}/{object_pk}/{site_id}",
    # The key 'comment_reactions' stores the json output produced by
    # InkComment.get_reactions(), for the comment receiving the method.
    "comment_reactions": "/comment_reactions/cm/{comment_id}",
    # The key 'object_reactions' stores the json output produced by
    # get_object_reactions(ctype_pk, object_pk, site_id).
    "object_reactions": "/object_reactions/{ctype_pk}/{object_pk}/{site_id}",
}



from __future__ import unicode_literals

from django.contrib import admin
from django_comments import get_model
from django_comments.admin import CommentsAdmin
from django_comments.models import CommentFlag
from django_comments_ink.models import BlackListedDomain, InkComment


class InkCommentsAdmin(CommentsAdmin):
    list_display = (
        "cid",
        "thread_level",
        "nested_count",
        "name",
        "content_type",
        "object_pk",
        "ip_address",
        "submit_date",
        "followup",
        "is_public",
        "is_removed",
    )
    list_display_links = ("cid",)
    list_filter = ("content_type", "is_public", "is_removed", "followup")
    fieldsets = (
        (None, {"fields": ("content_type", "object_pk", "site")}),
        (
            "Content",
            {
                "fields": (
                    "user",
                    "user_name",
                    "user_email",
                    "user_url",
                    "comment",
                    "followup",
                )
            },
        ),
        (
            "Metadata",
            {
                "fields": (
                    "submit_date",
                    "ip_address",
                    "is_public",
                    "is_removed",
                )
            },
        ),
    )
    date_hierarchy = "submit_date"
    ordering = ("thread_id", "order")
    search_fields = ["id", "object_pk", "user_name", "user_email", "comment"]

    def thread_level(self, obj):
        rep = "|"
        if obj.level:
            rep += "-" * obj.level
            rep += " c%d to c%d" % (obj.id, obj.parent_id)
        else:
            rep += " c%d" % obj.id
        return rep

    def cid(self, obj):
        return "c%d" % obj.id


class BlackListedDomainAdmin(admin.ModelAdmin):
    search_fields = ["domain"]


if get_model() is InkComment:
    admin.site.register(InkComment, InkCommentsAdmin)
    admin.site.register(CommentFlag)
    admin.site.register(BlackListedDomain, BlackListedDomainAdmin)



# Generated by Django 4.0.3 on 2022-03-29 20:26

import datetime

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ("django_comments_ink", "0001_initial"),
    ]

    operations = [
        migrations.CreateModel(
            name="Article",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "title",
                    models.CharField(max_length=200, verbose_name="title"),
                ),
                (
                    "slug",
                    models.SlugField(
                        unique_for_date="publish", verbose_name="slug"
                    ),
                ),
                ("body", models.TextField(verbose_name="body")),
                (
                    "allow_comments",
                    models.BooleanField(
                        default=True, verbose_name="allow comments"
                    ),
                ),
                (
                    "publish",
                    models.DateTimeField(
                        default=datetime.datetime.now, verbose_name="publish"
                    ),
                ),
            ],
            options={
                "db_table": "demo_articles",
                "ordering": ("-publish",),
            },
        ),
        migrations.CreateModel(
            name="Diary",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("body", models.TextField(verbose_name="body")),
                (
                    "allow_comments",
                    models.BooleanField(
                        default=True, verbose_name="allow comments"
                    ),
                ),
                (
                    "publish",
                    models.DateTimeField(
                        default=datetime.datetime.now, verbose_name="publish"
                    ),
                ),
            ],
            options={
                "db_table": "demo_diary",
                "ordering": ("-publish",),
            },
        ),
        migrations.CreateModel(
            name="MyComment",
            fields=[
                (
                    "inkcomment_ptr",
                    models.OneToOneField(
                        auto_created=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        parent_link=True,
                        primary_key=True,
                        serialize=False,
                        to="django_comments_ink.inkcomment",
                    ),
                ),
                ("title", models.CharField(max_length=256)),
            ],
            options={
                "verbose_name": "comment",
                "verbose_name_plural": "comments",
                "ordering": ("submit_date",),
                "permissions": [("can_moderate", "Can moderate comments")],
                "abstract": False,
            },
            bases=("django_comments_ink.inkcomment",),
        ),
    ]



from datetime import datetime
from unittest.mock import patch

import pytest
from django.contrib.contenttypes.models import ContentType
from django.contrib.sites.models import Site
from django.db.models.signals import pre_save
from django.test import TestCase as DjangoTestCase
from django_comments_ink import get_form, get_model
from django_comments_ink.models import (
    BlackListedDomain,
    InkComment,
    MaxThreadLevelExceededException,
    publish_or_withhold_on_pre_save,
)
from django_comments_ink.moderation import SpamModerator, moderator
from django_comments_ink.tests.models import Article, Diary, MyComment
from django_comments_ink.tests.test_views import post_article_comment


class ArticleBaseTestCase(DjangoTestCase):
    def setUp(self):
        self.article_1 = Article.objects.create(
            title="September", slug="september", body="During September..."
        )
        self.article_2 = Article.objects.create(
            title="October", slug="october", body="What I did on October..."
        )


class InkCommentManagerTestCase(ArticleBaseTestCase):
    def setUp(self):
        super(InkCommentManagerTestCase, self).setUp()
        self.article_ct = ContentType.objects.get(
            app_label="tests", model="article"
        )
        self.site1 = Site.objects.get(pk=1)
        self.site2 = Site.objects.create(domain="site2.com", name="site2.com")

    def post_comment_1(self):
        InkComment.objects.create(
            content_type=self.article_ct,
            object_pk=self.article_1.id,
            content_object=self.article_1,
            site=self.site1,
            comment="just a testing comment",
            submit_date=datetime.now(),
        )

    def post_comment_2(self):
        InkComment.objects.create(
            content_type=self.article_ct,
            object_pk=self.article_2.id,
            content_object=self.article_2,
            site=self.site1,
            comment="yet another comment",
            submit_date=datetime.now(),
        )

    def post_comment_3(self):
        InkComment.objects.create(
            content_type=self.article_ct,
            object_pk=self.article_2.id,
            content_object=self.article_2,
            site=self.site1,
            comment="and another one",
            submit_date=datetime.now(),
        )

    def post_comment_4(self):
        InkComment.objects.create(
            content_type=self.article_ct,
            object_pk=self.article_1.id,
            content_object=self.article_1,
            site=self.site2,
            comment="just a testing comment in site2",
            submit_date=datetime.now(),
        )

    def test_for_app_models(self):
        # there is no comment posted yet to article_1 nor article_2
        count = InkComment.objects.for_app_models("tests.article").count()
        self.assertEqual(count, 0)
        self.post_comment_1()
        count = InkComment.objects.for_app_models("tests.article").count()
        self.assertEqual(count, 1)
        self.post_comment_2()
        count = InkComment.objects.for_app_models("tests.article").count()
        self.assertEqual(count, 2)
        self.post_comment_3()
        count = InkComment.objects.for_app_models("tests.article").count()
        self.assertEqual(count, 3)
        self.post_comment_4()
        count = InkComment.objects.for_app_models("tests.article").count()
        self.assertEqual(count, 4)

    def test_multi_site_for_app_models(self):
        self.post_comment_1()  # To site1.
        self.post_comment_4()  # To site2.
        count_site1 = InkComment.objects.for_app_models(
            "tests.article", site=self.site1
        ).count()
        self.assertEqual(count_site1, 1)
        count_site2 = InkComment.objects.for_app_models(
            "tests.article", site=self.site2
        ).count()
        self.assertEqual(count_site2, 1)


# In order to test 'save' and '_calculate_thread_data' methods, simulate the
# following threads, in order of arrival:
#
# testcase cmt.id   parent level-0  level-1  level-2  level-3
#  step1     1        -      c1                               <-           c1
#  step1     2        -      c2                               <-           c2
#  step2     3        1      --       c3                      <-        c3.c1
#  step2     4        1      --       c4                      <-        c4.c1
#  step3     5        2      --       c5                      <-        c5.c2
#  step4     6        5      --       --        c6            <-     c6.c5.c2
#  step4     7        4      --       --        c7            <-     c7.c4.c1
#  step5     8        3      --       --        c8            <-     c8.c3.c1
#  step5     9        -      c9                               <-           c9
#  step6    10        7                                 c10   <- c10.c7.c4.c1
#  step6    11        8                                 c11   <- c11.c8.c3.c1


def thread_test_step_1(article, model=get_model(), **kwargs):
    article_ct = ContentType.objects.get(app_label="tests", model="article")
    if not "site" in kwargs:
        kwargs["site"] = Site.objects.get(pk=1)

    # post Comment 1 with parent_id 0
    model.objects.create(
        content_type=article_ct,
        object_pk=article.id,
        content_object=article,
        comment="c1",
        submit_date=datetime.now(),
        **kwargs,
    )

    # post Comment 2 with parent_id 0
    model.objects.create(
        content_type=article_ct,
        object_pk=article.id,
        content_object=article,
        comment="c2",
        submit_date=datetime.now(),
        **kwargs,
    )


def thread_test_step_2(article, model=get_model(), **kwargs):
    article_ct = ContentType.objects.get(app_label="tests", model="article")
    if not "site" in kwargs:
        kwargs["site"] = Site.objects.get(pk=1)
    if not "parent_id" in kwargs:
        kwargs["parent_id"] = 1

    # post Comment 3 to parent_id 1
    model.objects.create(
        content_type=article_ct,
        object_pk=article.id,
        content_object=article,
        comment="c3.c1",
        submit_date=datetime.now(),
        **kwargs,
    )

    # post Comment 4 to parent_id 1
    model.objects.create(
        content_type=article_ct,
        object_pk=article.id,
        content_object=article,
        comment="c4.c1",
        submit_date=datetime.now(),
        **kwargs,
    )


def thread_test_step_3(article, model=get_model(), **kwargs):
    article_ct = ContentType.objects.get(app_label="tests", model="article")
    if not "site" in kwargs:
        kwargs["site"] = Site.objects.get(pk=1)

    # post Comment 5 to parent_id 2
    model.objects.create(
        content_type=article_ct,
        object_pk=article.id,
        content_object=article,
        comment="5.c2",
        submit_date=datetime.now(),
        parent_id=2,
        **kwargs,
    )


def thread_test_step_4(article, model=get_model(), **kwargs):
    article_ct = ContentType.objects.get(app_label="tests", model="article")
    if not "site" in kwargs:
        kwargs["site"] = Site.objects.get(pk=1)

    # post Comment 6 to parent_id 5
    model.objects.create(
        content_type=article_ct,
        object_pk=article.id,
        content_object=article,
        comment="c6.c5.c2",
        submit_date=datetime.now(),
        parent_id=5,
        **kwargs,
    )

    # post Comment 7 to parent_id 4
    model.objects.create(
        content_type=article_ct,
        object_pk=article.id,
        content_object=article,
        comment="c7.c4.c1",
        submit_date=datetime.now(),
        parent_id=4,
        **kwargs,
    )


def thread_test_step_5(article, model=get_model(), **kwargs):
    article_ct = ContentType.objects.get(app_label="tests", model="article")
    if not "site" in kwargs:
        kwargs["site"] = Site.objects.get(pk=1)

    # post Comment 8 to parent_id 3
    model.objects.create(
        content_type=article_ct,
        object_pk=article.id,
        content_object=article,
        comment="c8.c3.c1",
        submit_date=datetime.now(),
        parent_id=3,
        **kwargs,
    )

    # post Comment 9 with parent_id 0
    model.objects.create(
        content_type=article_ct,
        object_pk=article.id,
        content_object=article,
        comment="c3",
        submit_date=datetime.now(),
        **kwargs,
    )


def thread_test_step_6(article, model=get_model(), **kwargs):
    article_ct = ContentType.objects.get(app_label="tests", model="article")
    if not "site" in kwargs:
        kwargs["site"] = Site.objects.get(pk=1)

    # post Comment 10 to parent_id 7
    model.objects.create(
        content_type=article_ct,
        object_pk=article.id,
        content_object=article,
        comment="c10.c7.c4.c1",
        submit_date=datetime.now(),
        parent_id=7,
        **kwargs,
    )

    # post Comment 11 to parent_id 8
    model.objects.create(
        content_type=article_ct,
        object_pk=article.id,
        content_object=article,
        comment="c11.c8.c3.c1",
        submit_date=datetime.now(),
        parent_id=8,
        **kwargs,
    )


class BaseThreadStep1TestCase(ArticleBaseTestCase):
    def setUp(self):
        super(BaseThreadStep1TestCase, self).setUp()
        thread_test_step_1(self.article_1)
        (  # content ->     cmt.id  thread_id  parent_id  level  order  nested
            self.c1,  # ->    1         1          1        0      1      0
            self.c2,  # ->    2         2          2        0      1      0
        ) = InkComment.objects.all()

    def test_threaded_comments_step_1_level_0(self):
        # comment 1
        self.assertTrue(self.c1.parent_id == 1 and self.c1.thread_id == 1)
        self.assertTrue(self.c1.level == 0 and self.c1.order == 1)
        self.assertEqual(self.c1.nested_count, 0)
        # comment 2
        self.assertTrue(self.c2.parent_id == 2 and self.c2.thread_id == 2)
        self.assertTrue(self.c2.level == 0 and self.c2.order == 1)
        self.assertEqual(self.c2.nested_count, 0)


class ThreadStep2TestCase(ArticleBaseTestCase):
    def setUp(self):
        super(ThreadStep2TestCase, self).setUp()
        thread_test_step_1(self.article_1)
        thread_test_step_2(self.article_1)
        (  # content ->    cmt.id  thread_id  parent_id  level  order  nested
            self.c1,  # ->   1         1          1        0      1      2
            self.c3,  # ->   3         1          1        1      2      0
            self.c4,  # ->   4         1          1        1      3      0
            self.c2,  # ->   2         2          2        0      1      0
        ) = InkComment.objects.all()

    def test_threaded_comments_step_2_level_0(self):
        # comment 1
        self.assertTrue(self.c1.parent_id == 1 and self.c1.thread_id == 1)
        self.assertTrue(self.c1.level == 0 and self.c1.order == 1)
        self.assertEqual(self.c1.nested_count, 2)
        # comment 2
        self.assertTrue(self.c2.parent_id == 2 and self.c2.thread_id == 2)
        self.assertTrue(self.c2.level == 0 and self.c2.order == 1)
        self.assertEqual(self.c2.nested_count, 0)

    def test_threaded_comments_step_2_level_1(self):
        # comment 3
        self.assertTrue(self.c3.parent_id == 1 and self.c3.thread_id == 1)
        self.assertTrue(self.c3.level == 1 and self.c3.order == 2)
        self.assertEqual(self.c3.nested_count, 0)
        # comment 4
        self.assertTrue(self.c4.parent_id == 1 and self.c4.thread_id == 1)
        self.assertTrue(self.c4.level == 1 and self.c4.order == 3)
        self.assertEqual(self.c4.nested_count, 0)


class ThreadStep3TestCase(ArticleBaseTestCase):
    def setUp(self):
        super(ThreadStep3TestCase, self).setUp()
        thread_test_step_1(self.article_1)
        thread_test_step_2(self.article_1)
        thread_test_step_3(self.article_1)

        (  # -> content:   cmt.id  thread_id  parent_id  level  order  nested
            self.c1,  # ->   1         1          1        0      1      2
            self.c3,  # ->   3         1          1        1      2      0
            self.c4,  # ->   4         1          1        1      3      0
            self.c2,  # ->   2         2          2        0      1      1
            self.c5,  # ->   5         2          2        1      2      0
        ) = InkComment.objects.all()

    def test_threaded_comments_step_3_level_0(self):
        # comment 1
        self.assertTrue(self.c1.parent_id == 1 and self.c1.thread_id == 1)
        self.assertTrue(self.c1.level == 0 and self.c1.order == 1)
        self.assertEqual(self.c1.nested_count, 2)
        # comment 2
        self.assertTrue(self.c2.parent_id == 2 and self.c2.thread_id == 2)
        self.assertTrue(self.c2.level == 0 and self.c2.order == 1)
        self.assertEqual(self.c2.nested_count, 1)

    def test_threaded_comments_step_3_level_1(self):
        # comment 3
        self.assertTrue(self.c3.parent_id == 1 and self.c3.thread_id == 1)
        self.assertTrue(self.c3.level == 1 and self.c3.order == 2)
        self.assertEqual(self.c3.nested_count, 0)
        # comment 4
        self.assertTrue(self.c4.parent_id == 1 and self.c4.thread_id == 1)
        self.assertTrue(self.c4.level == 1 and self.c4.order == 3)
        self.assertEqual(self.c4.nested_count, 0)
        # comment 5
        self.assertTrue(self.c5.parent_id == 2 and self.c5.thread_id == 2)
        self.assertTrue(self.c5.level == 1 and self.c5.order == 2)
        self.assertEqual(self.c5.nested_count, 0)


class ThreadStep4TestCase(ArticleBaseTestCase):
    def setUp(self):
        super(ThreadStep4TestCase, self).setUp()
        thread_test_step_1(self.article_1)
        thread_test_step_2(self.article_1)
        thread_test_step_3(self.article_1)
        thread_test_step_4(self.article_1)

        (  # content ->    cmt.id  thread_id  parent_id  level  order  nested
            self.c1,  # ->   1         1          1        0      1      3
            self.c3,  # ->   3         1          1        1      2      0
            self.c4,  # ->   4         1          1        1      3      1
            self.c7,  # ->   7         1          4        2      4      0
            self.c2,  # ->   2         2          2        0      1      2
            self.c5,  # ->   5         2          2        1      2      1
            self.c6,  # ->   6         2          5        2      3      0
        ) = InkComment.objects.all()

    def test_threaded_comments_step_4_level_0(self):
        # comment 1
        self.assertTrue(self.c1.parent_id == 1 and self.c1.thread_id == 1)
        self.assertTrue(self.c1.level == 0 and self.c1.order == 1)
        self.assertEqual(self.c1.nested_count, 3)
        # comment 2
        self.assertTrue(self.c2.parent_id == 2 and self.c2.thread_id == 2)
        self.assertTrue(self.c2.level == 0 and self.c2.order == 1)
        self.assertEqual(self.c2.nested_count, 2)

    def test_threaded_comments_step_4_level_1(self):
        # comment 3
        self.assertTrue(self.c3.parent_id == 1 and self.c3.thread_id == 1)
        self.assertTrue(self.c3.level == 1 and self.c3.order == 2)
        self.assertEqual(self.c3.nested_count, 0)
        # comment 4
        self.assertTrue(self.c4.parent_id == 1 and self.c4.thread_id == 1)
        self.assertTrue(self.c4.level == 1 and self.c4.order == 3)
        self.assertEqual(self.c4.nested_count, 1)
        # comment 5
        self.assertTrue(self.c5.parent_id == 2 and self.c5.thread_id == 2)
        self.assertTrue(self.c5.level == 1 and self.c5.order == 2)
        self.assertEqual(self.c5.nested_count, 1)

    def test_threaded_comments_step_4_level_2(self):
        # comment 6
        self.assertTrue(self.c6.parent_id == 5 and self.c6.thread_id == 2)
        self.assertTrue(self.c6.level == 2 and self.c6.order == 3)
        self.assertEqual(self.c6.nested_count, 0)
        # comment 7
        self.assertTrue(self.c7.parent_id == 4 and self.c7.thread_id == 1)
        self.assertTrue(self.c7.level == 2 and self.c7.order == 4)
        self.assertEqual(self.c7.nested_count, 0)


class ThreadStep5TestCase(ArticleBaseTestCase):
    def setUp(self):
        super(ThreadStep5TestCase, self).setUp()
        thread_test_step_1(self.article_1)
        thread_test_step_2(self.article_1)
        thread_test_step_3(self.article_1)
        thread_test_step_4(self.article_1)
        thread_test_step_5(self.article_1)

        (  # content ->    cmt.id  thread_id  parent_id  level  order  nested
            self.c1,  # ->   1         1          1        0      1      4
            self.c3,  # ->   |- 3      1          1        1      2      1
            self.c8,  # ->      |- 8   1          3        2      3      0
            self.c4,  # ->   |- 4      1          1        1      4      1
            self.c7,  # ->      |- 7   1          4        2      5      0
            self.c2,  # ->   2         2          2        0      1      2
            self.c5,  # ->   |- 5      2          2        1      2      1
            self.c6,  # ->      |- 6   2          5        2      3      0
            self.c9,  # ->   9         9          9        0      1      0
        ) = InkComment.objects.all()

    def test_threaded_comments_step_5_level_0(self):
        # comment 1
        self.assertTrue(self.c1.parent_id == 1 and self.c1.thread_id == 1)
        self.assertTrue(self.c1.level == 0 and self.c1.order == 1)
        self.assertEqual(self.c1.nested_count, 4)
        # comment 2
        self.assertTrue(self.c2.parent_id == 2 and self.c2.thread_id == 2)
        self.assertTrue(self.c2.level == 0 and self.c2.order == 1)
        self.assertEqual(self.c2.nested_count, 2)
        # comment 9
        self.assertTrue(self.c9.parent_id == 9 and self.c9.thread_id == 9)
        self.assertTrue(self.c9.level == 0 and self.c9.order == 1)
        self.assertEqual(self.c9.nested_count, 0)

    def test_threaded_comments_step_5_level_1(self):
        # comment 3
        self.assertTrue(self.c3.parent_id == 1 and self.c3.thread_id == 1)
        self.assertTrue(self.c3.level == 1 and self.c3.order == 2)
        self.assertEqual(self.c3.nested_count, 1)
        # comment 4
        self.assertTrue(self.c4.parent_id == 1 and self.c4.thread_id == 1)
        self.assertTrue(self.c4.level == 1 and self.c4.order == 4)  # changed
        self.assertEqual(self.c4.nested_count, 1)
        # comment 5
        self.assertTrue(self.c5.parent_id == 2 and self.c5.thread_id == 2)
        self.assertTrue(self.c5.level == 1 and self.c5.order == 2)
        self.assertEqual(self.c5.nested_count, 1)

    def test_threaded_comments_step_5_level_2(self):
        # comment 6
        self.assertTrue(self.c6.parent_id == 5 and self.c6.thread_id == 2)
        self.assertTrue(self.c6.level == 2 and self.c6.order == 3)
        self.assertEqual(self.c6.nested_count, 0)
        # comment 7
        self.assertTrue(self.c7.parent_id == 4 and self.c7.thread_id == 1)
        self.assertTrue(self.c7.level == 2 and self.c7.order == 5)  # changed
        self.assertEqual(self.c7.nested_count, 0)
        # comment 8
        self.assertTrue(self.c8.parent_id == 3 and self.c8.thread_id == 1)
        self.assertTrue(self.c8.level == 2 and self.c8.order == 3)
        self.assertEqual(self.c8.nested_count, 0)

    @patch.multiple(
        "django_comments_ink.conf.settings", COMMENTS_INK_MAX_THREAD_LEVEL=2
    )
    def test_exceed_max_thread_level_raises_exception(self):
        article_ct = ContentType.objects.get(app_label="tests", model="article")
        site = Site.objects.get(pk=1)
        with self.assertRaises(MaxThreadLevelExceededException):
            InkComment.objects.create(
                content_type=article_ct,
                object_pk=self.article_1.id,
                content_object=self.article_1,
                site=site,
                comment="cmt 1 to cmt 2 to cmt 1",
                submit_date=datetime.now(),
                parent_id=8,
            )  # already max thread level

    def test_removing_c4_withdraws_c7_and_updates_nested_count(self):
        cm4 = InkComment.objects.get(pk=4)
        self.assertEqual(cm4.nested_count, 1)
        cm1 = InkComment.objects.get(pk=1)
        self.assertEqual(cm1.nested_count, 4)
        # Remove comment 4, save, and check again.
        cm4.is_removed = True
        cm4.save()
        cm4 = InkComment.objects.get(pk=4)
        self.assertEqual(cm4.nested_count, 1)
        cm1 = InkComment.objects.get(pk=1)
        self.assertEqual(cm1.nested_count, 3)


class ThreadStep6TestCase(ArticleBaseTestCase):
    def setUp(self):
        super(ThreadStep6TestCase, self).setUp()
        thread_test_step_1(self.article_1)
        thread_test_step_2(self.article_1)
        thread_test_step_3(self.article_1)
        thread_test_step_4(self.article_1)
        thread_test_step_5(self.article_1)
        thread_test_step_6(self.article_1)

        (  # content ->    cmt.id  thread_id  parent_id  level  order  nested
            self.c1,  # ->   1         1          1        0      1      6
            self.c3,  # ->   3         1          1        1      2      2
            self.c8,  # ->   8         1          3        2      3      1
            self.c11,  # ->  11        1          8        3      4      0
            self.c4,  # ->   4         1          1        1      5      2
            self.c7,  # ->   7         1          4        2      6      1
            self.c10,  # ->  10        1          7        3      7      0
            self.c2,  # ->   2         2          2        0      1      2
            self.c5,  # ->   5         2          2        1      2      1
            self.c6,  # ->   6         2          5        2      3      0
            self.c9,  # ->   9         9          9        0      1      0
        ) = InkComment.objects.all()

    def test_threaded_comments_step_6_level_0(self):
        # comment 1
        self.assertTrue(self.c1.parent_id == 1 and self.c1.thread_id == 1)
        self.assertTrue(self.c1.level == 0 and self.c1.order == 1)
        self.assertEqual(self.c1.nested_count, 6)
        # comment 2
        self.assertTrue(self.c2.parent_id == 2 and self.c2.thread_id == 2)
        self.assertTrue(self.c2.level == 0 and self.c2.order == 1)
        self.assertEqual(self.c2.nested_count, 2)
        # comment 9
        self.assertTrue(self.c9.parent_id == 9 and self.c9.thread_id == 9)
        self.assertTrue(self.c9.level == 0 and self.c9.order == 1)
        self.assertEqual(self.c9.nested_count, 0)

    def test_threaded_comments_step_6_level_1(self):
        # comment 3
        self.assertTrue(self.c3.parent_id == 1 and self.c3.thread_id == 1)
        self.assertTrue(self.c3.level == 1 and self.c3.order == 2)
        self.assertEqual(self.c3.nested_count, 2)
        # comment 4
        self.assertTrue(self.c4.parent_id == 1 and self.c4.thread_id == 1)
        self.assertTrue(self.c4.level == 1 and self.c4.order == 5)
        self.assertEqual(self.c4.nested_count, 2)
        # comment 5
        self.assertTrue(self.c5.parent_id == 2 and self.c5.thread_id == 2)
        self.assertTrue(self.c5.level == 1 and self.c5.order == 2)
        self.assertEqual(self.c5.nested_count, 1)

    def test_threaded_comments_step_6_level_2(self):
        # comment 8
        self.assertTrue(self.c8.parent_id == 3 and self.c8.thread_id == 1)
        self.assertTrue(self.c8.level == 2 and self.c8.order == 3)
        self.assertEqual(self.c8.nested_count, 1)
        # comment 7
        self.assertTrue(self.c7.parent_id == 4 and self.c7.thread_id == 1)
        self.assertTrue(self.c7.level == 2 and self.c7.order == 6)
        self.assertEqual(self.c7.nested_count, 1)
        # comment 6
        self.assertTrue(self.c6.parent_id == 5 and self.c6.thread_id == 2)
        self.assertTrue(self.c6.level == 2 and self.c6.order == 3)
        self.assertEqual(self.c6.nested_count, 0)

    def test_threaded_comments_step_6_level_3(self):
        # comment 10
        self.assertTrue(self.c10.parent_id == 7 and self.c10.thread_id == 1)
        self.assertTrue(self.c10.level == 3 and self.c10.order == 7)
        self.assertEqual(self.c10.nested_count, 0)
        # comment 11
        self.assertTrue(self.c11.parent_id == 8 and self.c11.thread_id == 1)
        self.assertTrue(self.c11.level == 3 and self.c11.order == 4)
        self.assertEqual(self.c11.nested_count, 0)


def add_comment_to_diary_entry(diary_entry):
    diary_ct = ContentType.objects.get(app_label="tests", model="diary")
    site = Site.objects.get(pk=1)
    get_model().objects.create(
        content_type=diary_ct,
        object_pk=diary_entry.id,
        content_object=diary_entry,
        site=site,
        comment="cmt to day in diary",
        submit_date=datetime.now(),
    )


class DiaryBaseTestCase(DjangoTestCase):
    def setUp(self):
        self.day_in_diary = Diary.objects.create(body="About Today...")
        add_comment_to_diary_entry(self.day_in_diary)

    def test_max_thread_level_by_app_model(self):
        diary_ct = ContentType.objects.get(app_label="tests", model="diary")
        site = Site.objects.get(pk=1)
        with self.assertRaises(MaxThreadLevelExceededException):
            InkComment.objects.create(
                content_type=diary_ct,
                object_pk=self.day_in_diary.id,
                content_object=self.day_in_diary,
                site=site,
                comment="cmt to cmt to day in diary",
                submit_date=datetime.now(),
                parent_id=1,
            )  # already max thread level


class PublishOrWithholdNestedComments_1_TestCase(ArticleBaseTestCase):
    # Add a threaded comment structure (c1, c2, c3) and verify that
    # removing c1 withholds c3.

    def setUp(self):
        super(PublishOrWithholdNestedComments_1_TestCase, self).setUp()
        thread_test_step_1(self.article_1)
        thread_test_step_2(self.article_1)
        #
        # These two lines create the following comments:
        #
        # (  # content ->    cmt.id  thread_id  parent_id  level  order  nested
        #     cm1,   # ->     1         1          1        0      1       2
        #     cm3,   # ->     3         1          1        1      2       0
        #     cm4,   # ->     4         1          1        1      3       0
        #     cm2,   # ->     2         2          2        0      1       0
        # ) = InkComment.objects.all()

    def test_all_comments_are_public_and_have_not_been_removed(self):
        for cm in InkComment.objects.all():
            self.assertTrue(cm.is_public)
            self.assertFalse(cm.is_removed)

    def test_removing_c1_withholds_c3_and_c4(self):
        cm1 = InkComment.objects.get(pk=1)
        self.assertEqual(cm1.nested_count, 2)  # nested_count should be 2.

        cm1.is_removed = True
        cm1.save()
        cm1 = InkComment.objects.get(pk=1)
        self.assertTrue(cm1.is_public)
        self.assertTrue(cm1.is_removed)
        # Is still public, so the nested_count doesn't change.
        self.assertEqual(cm1.nested_count, 2)

        cm3 = InkComment.objects.get(pk=3)
        self.assertFalse(cm3.is_public)
        self.assertFalse(cm3.is_removed)

        cm4 = InkComment.objects.get(pk=4)
        self.assertFalse(cm4.is_public)
        self.assertFalse(cm4.is_removed)


_ink_model = "django_comments_ink.tests.models.MyComment"


class PublishOrWithholdNestedComments_2_TestCase(ArticleBaseTestCase):
    # Then mock the settings so that the project uses a customized
    # comment model (django_comments_ink.tests.models.MyComment), and repeat
    # the logic adding MyComment instances. Then remove c1 and be sure
    # that c3 gets withheld.

    def setUp(self):
        super(PublishOrWithholdNestedComments_2_TestCase, self).setUp()
        thread_test_step_1(
            self.article_1, model=MyComment, title="Can't be empty 1"
        )
        thread_test_step_2(
            self.article_1, model=MyComment, title="Can't be empty 2"
        )
        #
        # These two lines create the following comments:
        #
        # (  # content ->    cmt.id thread_id parent_id level order nested
        #     cm1,   # ->     1        1         1        0     1      2
        #     cm3,   # ->     3        1         1        1     2      0
        #     cm4,   # ->     4        1         1        1     3      0
        #     cm2,   # ->     2        2         2        0     1      0
        # ) = MyComment.objects.all()

    def test_all_comments_are_public_and_have_not_been_removed(self):
        for cm in MyComment.objects.all():
            self.assertTrue(cm.is_public)
            self.assertFalse(cm.is_removed)

    @patch.multiple(
        "django_comments_ink.conf.settings", COMMENTS_INK_MODEL=_ink_model
    )
    def test_removing_c1_withholds_c3_and_c4(self):
        # Register the receiver again. It was registered in apps.py, but we
        # have patched the COMMENTS_INK_MODEL, however we won't fake the
        # ready. It's easier to just register again the receiver, to test
        # only what depends on django-comments-ink.
        model_app_label = get_model()._meta.label
        pre_save.connect(
            publish_or_withhold_on_pre_save, sender=model_app_label
        )

        cm1 = MyComment.objects.get(pk=1)
        cm1.is_removed = True
        cm1.save()
        self.assertTrue(cm1.is_public)
        self.assertTrue(cm1.is_removed)

        cm3 = MyComment.objects.get(pk=3)
        self.assertFalse(cm3.is_public)
        self.assertFalse(cm3.is_removed)

        cm4 = MyComment.objects.get(pk=4)
        self.assertFalse(cm4.is_public)
        self.assertFalse(cm4.is_removed)


@pytest.mark.django_db
def test_get_reply_url(an_articles_comment):
    reply_url = an_articles_comment.get_reply_url()
    assert reply_url == f"/comments/reply/{an_articles_comment.pk}/"


@pytest.mark.django_db
def test_get_queryset_returns_none():
    qs = InkComment.get_queryset(
        content_type=None, object_pk=None, content_object=None
    )
    assert qs is None


@pytest.mark.django_db
def test_get_queryset_with_content_object(an_article, an_articles_comment):
    qs = InkComment.get_queryset(content_object=an_article)
    assert qs[0] == an_articles_comment


# ---------------------------------------------------------------------
# Test BlackListedDomain.


class ArticleCommentModerator(SpamModerator):
    pass


@pytest.mark.django_db
def test_blacklisted_domain_is_blocked(an_article, an_user):
    domain = BlackListedDomain.objects.create(domain="example.com")
    moderator.register(Article, ArticleCommentModerator)
    form = get_form()(an_article)
    data = {
        "name": "Joe",
        "email": f"joe@{domain}",
        "followup": True,
        "reply_to": 0,
        "level": 1,
        "order": 1,
        "comment": "Es war einmal eine kleine...",
    }
    data.update(form.initial)
    response = post_article_comment(data, an_article, auth_user=an_user)
    assert response.status_code == 400  # It would be 302 otherwise.
    moderator.unregister(Article)
    assert InkComment.objects.count() == 0


@pytest.mark.django_db
def test_non_blacklisted_domain_pass(an_article, an_user):
    moderator.register(Article, ArticleCommentModerator)
    form = get_form()(an_article)
    data = {
        "name": "Joe",
        "email": f"joe@example.com",
        "followup": True,
        "reply_to": 0,
        "level": 1,
        "order": 1,
        "comment": "Es war einmal eine kleine...",
    }
    data.update(form.initial)
    response = post_article_comment(data, an_article, auth_user=an_user)
    assert response.status_code == 302
    moderator.unregister(Article)
    assert InkComment.objects.count() == 1


# ---------------------------------------------------------------------
# Test django.db.models.signals.post_delete signal.


@pytest.mark.django_db
def test_nested_count_after_deleting_comment_1(an_article):
    thread_test_step_1(an_article)
    thread_test_step_2(an_article)
    thread_test_step_3(an_article)
    thread_test_step_4(an_article)
    thread_test_step_5(an_article)
    thread_test_step_6(an_article)

    # content -> cmt.id  thread_id  parent_id  level  order  nested
    #  c1   # ->    1         1          1        0      1      6
    #  c3   # ->    3         1          1        1      2      2
    #  c8   # ->    8         1          3        2      3      1
    #  c11  # ->   11         1          8        3      4      0
    #  c4   # ->    4         1          1        1      5      2
    #  c7   # ->    7         1          4        2      6      1
    #  c10  # ->   10         1          7        3      7      0
    #  c2   # ->    2         2          2        0      1      2
    #  c5   # ->    5         2          2        1      2      1
    #  c6   # ->    6         2          5        2      3      0
    #  c9   # ->    9         9          9        0      1      0

    cm1 = InkComment.norel_objects.get(pk=1)
    cm1.delete()

    # It should remove comments 1, 3, 8, 11, 4, 7 and 10.
    # As the comment deleted was at level 0, there is no nested_count
    # record to modify.

    for cid in [1, 3, 8, 11, 4, 7, 10]:
        with pytest.raises(InkComment.DoesNotExist):
            InkComment.objects.get(pk=cid)


@pytest.mark.django_db
def test_nested_count_after_deleting_comment_2(an_article):
    thread_test_step_1(an_article)
    thread_test_step_2(an_article)
    thread_test_step_3(an_article)
    thread_test_step_4(an_article)
    thread_test_step_5(an_article)
    thread_test_step_6(an_article)

    # content -> cmt.id  thread_id  parent_id  level  order  nested
    #  c1   # ->    1         1          1        0      1      6
    #  c3   # ->    3         1          1        1      2      2
    #  c8   # ->    8         1          3        2      3      1
    #  c11  # ->   11         1          8        3      4      0
    #  c4   # ->    4         1          1        1      5      2
    #  c7   # ->    7         1          4        2      6      1
    #  c10  # ->   10         1          7        3      7      0
    #  c2   # ->    2         2          2        0      1      2
    #  c5   # ->    5         2          2        1      2      1
    #  c6   # ->    6         2          5        2      3      0
    #  c9   # ->    9         9          9        0      1      0

    cm2 = InkComment.norel_objects.get(pk=2)
    cm2.delete()

    # It should remove comments 2, 5 and 6.
    # As the comment deleted was at level 0, there is no nested_count
    # record to modify.

    for cid in [2, 5, 6]:
        with pytest.raises(InkComment.DoesNotExist):
            InkComment.objects.get(pk=cid)


@pytest.mark.django_db
def test_nested_count_after_deleting_comment_3(an_article):
    thread_test_step_1(an_article)
    thread_test_step_2(an_article)
    thread_test_step_3(an_article)
    thread_test_step_4(an_article)
    thread_test_step_5(an_article)
    thread_test_step_6(an_article)

    # content -> cmt.id  thread_id  parent_id  level  order  nested
    #  c1   # ->    1         1          1        0      1      6
    #  c3   # ->    3         1          1        1      2      2
    #  c8   # ->    8         1          3        2      3      1
    #  c11  # ->   11         1          8        3      4      0
    #  c4   # ->    4         1          1        1      5      2
    #  c7   # ->    7         1          4        2      6      1
    #  c10  # ->   10         1          7        3      7      0
    #  c2   # ->    2         2          2        0      1      2
    #  c5   # ->    5         2          2        1      2      1
    #  c6   # ->    6         2          5        2      3      0
    #  c9   # ->    9         9          9        0      1      0

    cm3 = InkComment.norel_objects.get(pk=3)
    cm3.delete()

    # It should remove comments 3, 8 and 11, and leave the following changes:
    # content -> cmt.id  thread_id  parent_id  level  order  nested
    #  c1   # ->    1         1          1        0      1      3

    for cid in [3, 8, 11]:
        with pytest.raises(InkComment.DoesNotExist):
            InkComment.objects.get(pk=cid)

    c1 = InkComment.norel_objects.get(pk=1)
    assert c1.nested_count == 3


@pytest.mark.django_db
def test_nested_count_after_deleting_comment_4(an_article):
    thread_test_step_1(an_article)
    thread_test_step_2(an_article)
    thread_test_step_3(an_article)
    thread_test_step_4(an_article)
    thread_test_step_5(an_article)
    thread_test_step_6(an_article)

    # content -> cmt.id  thread_id  parent_id  level  order  nested
    #  c1   # ->    1         1          1        0      1      6
    #  c3   # ->    3         1          1        1      2      2
    #  c8   # ->    8         1          3        2      3      1
    #  c11  # ->   11         1          8        3      4      0
    #  c4   # ->    4         1          1        1      5      2
    #  c7   # ->    7         1          4        2      6      1
    #  c10  # ->   10         1          7        3      7      0
    #  c2   # ->    2         2          2        0      1      2
    #  c5   # ->    5         2          2        1      2      1
    #  c6   # ->    6         2          5        2      3      0
    #  c9   # ->    9         9          9        0      1      0

    cm4 = InkComment.norel_objects.get(pk=4)
    cm4.delete()

    # It should remove comments 4, 7 and 10, and leave the following changes:
    # content -> cmt.id  thread_id  parent_id  level  order  nested
    #  c1   # ->    1         1          1        0      1      3

    for cid in [4, 7, 10]:
        with pytest.raises(InkComment.DoesNotExist):
            InkComment.objects.get(pk=cid)

    c1 = InkComment.norel_objects.get(pk=1)
    assert c1.nested_count == 3


@pytest.mark.django_db
def test_nested_count_after_deleting_comment_5(an_article):
    thread_test_step_1(an_article)
    thread_test_step_2(an_article)
    thread_test_step_3(an_article)
    thread_test_step_4(an_article)
    thread_test_step_5(an_article)
    thread_test_step_6(an_article)

    # content -> cmt.id  thread_id  parent_id  level  order  nested
    #  c1   # ->    1         1          1        0      1      6
    #  c3   # ->    3         1          1        1      2      2
    #  c8   # ->    8         1          3        2      3      1
    #  c11  # ->   11         1          8        3      4      0
    #  c4   # ->    4         1          1        1      5      2
    #  c7   # ->    7         1          4        2      6      1
    #  c10  # ->   10         1          7        3      7      0
    #  c2   # ->    2         2          2        0      1      2
    #  c5   # ->    5         2          2        1      2      1
    #  c6   # ->    6         2          5        2      3      0
    #  c9   # ->    9         9          9        0      1      0

    cm5 = InkComment.norel_objects.get(pk=5)
    cm5.delete()

    # It should remove comments 5 and 6, and leave the following changes:
    # content -> cmt.id  thread_id  parent_id  level  order  nested
    #  c2   # ->    2         2          2        0      1      0

    for cid in [5, 6]:
        with pytest.raises(InkComment.DoesNotExist):
            InkComment.objects.get(pk=cid)

    c2 = InkComment.norel_objects.get(pk=2)
    assert c2.nested_count == 0


@pytest.mark.django_db
def test_nested_count_after_deleting_comment_6(an_article):
    thread_test_step_1(an_article)
    thread_test_step_2(an_article)
    thread_test_step_3(an_article)
    thread_test_step_4(an_article)
    thread_test_step_5(an_article)
    thread_test_step_6(an_article)

    # content -> cmt.id  thread_id  parent_id  level  order  nested
    #  c1   # ->    1         1          1        0      1      6
    #  c3   # ->    3         1          1        1      2      2
    #  c8   # ->    8         1          3        2      3      1
    #  c11  # ->   11         1          8        3      4      0
    #  c4   # ->    4         1          1        1      5      2
    #  c7   # ->    7         1          4        2      6      1
    #  c10  # ->   10         1          7        3      7      0
    #  c2   # ->    2         2          2        0      1      2
    #  c5   # ->    5         2          2        1      2      1
    #  c6   # ->    6         2          5        2      3      0
    #  c9   # ->    9         9          9        0      1      0

    cm6 = InkComment.norel_objects.get(pk=6)
    cm6.delete()

    # It should remove comment 6, and leave the following changes:
    # content -> cmt.id  thread_id  parent_id  level  order  nested
    #  c2   # ->    2         2          2        0      1      1
    #  c5   # ->    5         2          2        1      2      0

    with pytest.raises(InkComment.DoesNotExist):
        InkComment.objects.get(pk=6)

    c2 = InkComment.norel_objects.get(pk=2)
    assert c2.nested_count == 1

    c5 = InkComment.norel_objects.get(pk=5)
    assert c5.nested_count == 0


@pytest.mark.django_db
def test_nested_count_after_deleting_comment_7(an_article):
    thread_test_step_1(an_article)
    thread_test_step_2(an_article)
    thread_test_step_3(an_article)
    thread_test_step_4(an_article)
    thread_test_step_5(an_article)
    thread_test_step_6(an_article)

    # content -> cmt.id  thread_id  parent_id  level  order  nested
    #  c1   # ->    1         1          1        0      1      6
    #  c3   # ->    3         1          1        1      2      2
    #  c8   # ->    8         1          3        2      3      1
    #  c11  # ->   11         1          8        3      4      0
    #  c4   # ->    4         1          1        1      5      2
    #  c7   # ->    7         1          4        2      6      1
    #  c10  # ->   10         1          7        3      7      0
    #  c2   # ->    2         2          2        0      1      2
    #  c5   # ->    5         2          2        1      2      1
    #  c6   # ->    6         2          5        2      3      0
    #  c9   # ->    9         9          9        0      1      0

    cm7 = InkComment.norel_objects.get(pk=7)
    cm7.delete()

    # It should remove comments 7 and 10, and leave the following changes:
    # content -> cmt.id  thread_id  parent_id  level  order  nested
    #  c1   # ->    1         1          1        0      1      4
    #  c4   # ->    4         1          1        1      5      0

    for cid in [7, 10]:
        with pytest.raises(InkComment.DoesNotExist):
            InkComment.objects.get(pk=cid)

    c1 = InkComment.norel_objects.get(pk=1)
    assert c1.nested_count == 4

    c4 = InkComment.norel_objects.get(pk=4)
    assert c4.nested_count == 0


@pytest.mark.django_db
def test_nested_count_after_deleting_comment_8(an_article):
    thread_test_step_1(an_article)
    thread_test_step_2(an_article)
    thread_test_step_3(an_article)
    thread_test_step_4(an_article)
    thread_test_step_5(an_article)
    thread_test_step_6(an_article)

    # content -> cmt.id  thread_id  parent_id  level  order  nested
    #  c1   # ->    1         1          1        0      1      6
    #  c3   # ->    3         1          1        1      2      2
    #  c8   # ->    8         1          3        2      3      1
    #  c11  # ->   11         1          8        3      4      0
    #  c4   # ->    4         1          1        1      5      2
    #  c7   # ->    7         1          4        2      6      1
    #  c10  # ->   10         1          7        3      7      0
    #  c2   # ->    2         2          2        0      1      2
    #  c5   # ->    5         2          2        1      2      1
    #  c6   # ->    6         2          5        2      3      0
    #  c9   # ->    9         9          9        0      1      0

    cm8 = InkComment.norel_objects.get(pk=8)
    cm8.delete()

    # It should remove comments 8 and 11, and leave the following changes:
    # content -> cmt.id  thread_id  parent_id  level  order  nested
    #  c1   # ->    1         1          1        0      1      4
    #  c3   # ->    3         1          1        1      2      0

    for cid in [8, 11]:
        with pytest.raises(InkComment.DoesNotExist):
            InkComment.objects.get(pk=cid)

    c1 = InkComment.norel_objects.get(pk=1)
    assert c1.nested_count == 4

    c3 = InkComment.norel_objects.get(pk=3)
    assert c3.nested_count == 0


@pytest.mark.django_db
def test_nested_count_after_deleting_comment_10(an_article):
    thread_test_step_1(an_article)
    thread_test_step_2(an_article)
    thread_test_step_3(an_article)
    thread_test_step_4(an_article)
    thread_test_step_5(an_article)
    thread_test_step_6(an_article)

    # content -> cmt.id  thread_id  parent_id  level  order  nested
    #  c1   # ->    1         1          1        0      1      6
    #  c3   # ->    3         1          1        1      2      2
    #  c8   # ->    8         1          3        2      3      1
    #  c11  # ->   11         1          8        3      4      0
    #  c4   # ->    4         1          1        1      5      2
    #  c7   # ->    7         1          4        2      6      1
    #  c10  # ->   10         1          7        3      7      0
    #  c2   # ->    2         2          2        0      1      2
    #  c5   # ->    5         2          2        1      2      1
    #  c6   # ->    6         2          5        2      3      0
    #  c9   # ->    9         9          9        0      1      0

    cm10 = InkComment.norel_objects.get(pk=10)
    cm10.delete()

    # It should remove comments 8 and 11, and leave the following changes:
    # content -> cmt.id  thread_id  parent_id  level  order  nested
    #  c1   # ->    1         1          1        0      1      5
    #  c4   # ->    4         1          1        1      5      1
    #  c7   # ->    7         1          4        2      6      0

    with pytest.raises(InkComment.DoesNotExist):
        InkComment.objects.get(pk=10)

    c1 = InkComment.norel_objects.get(pk=1)
    assert c1.nested_count == 5

    c4 = InkComment.norel_objects.get(pk=4)
    assert c4.nested_count == 1

    c7 = InkComment.norel_objects.get(pk=7)
    assert c7.nested_count == 0



from django.contrib.contenttypes.models import ContentType
from django.db.models import F
from django.shortcuts import get_object_or_404
from django_comments.views.moderation import perform_flag
from django_comments_ink import get_model as get_comment_model
from django_comments_ink.api import serializers
from django_comments_ink.conf import settings
from django_comments_ink.models import CommentReaction
from django_comments_ink.utils import check_option, get_current_site_id
from rest_framework import generics, mixins, permissions, renderers, status
from rest_framework.response import Response
from rest_framework.schemas.openapi import AutoSchema

InkComment = get_comment_model()


class DefaultsMixin:
    @property
    def renderer_classes(self):
        if self.kwargs.get("override_drf_defaults", False):
            return (renderers.JSONRenderer, renderers.BrowsableAPIRenderer)
        return super().renderer_classes

    @property
    def pagination_class(self):
        if self.kwargs.get("override_drf_defaults", False):
            return None
        return super().pagination_class


class CommentCreate(DefaultsMixin, generics.CreateAPIView):
    """Create a comment."""

    serializer_class = serializers.WriteCommentSerializer

    def post(self, request, *args, **kwargs):
        serializer = self.get_serializer(data=request.data)
        if serializer.is_valid():
            response = super(CommentCreate, self).post(request, *args, **kwargs)
        else:
            if "non_field_errors" in serializer.errors:
                response_msg = serializer.errors["non_field_errors"][0]
            else:
                response_msg = [k for k in serializer.errors.keys()]
            return Response(response_msg, status=400)
        if self.resp_dict["code"] == 201:  # The comment has been created.
            response.data.update(
                {"id": self.resp_dict["comment"]["ink_comment"].id}
            )
            return response
        elif self.resp_dict["code"] in [202, 204, 403]:
            return Response({}, status=self.resp_dict["code"])

    def perform_create(self, serializer):
        self.resp_dict = serializer.save()


class CommentList(DefaultsMixin, generics.ListAPIView):
    """List all comments for a given ContentType and object ID."""

    serializer_class = serializers.ReadCommentSerializer
    permission_classes = (permissions.AllowAny,)

    def get_queryset(self, **kwargs):
        content_type_arg = self.kwargs.get("content_type", None)
        object_pk_arg = self.kwargs.get("object_pk", None)
        app, model = content_type_arg.split("-")
        site_id = get_current_site_id(self.request)
        try:
            content_type = ContentType.objects.get_by_natural_key(app, model)
        except ContentType.DoesNotExist:
            return InkComment.objects.none()
        else:
            return InkComment.get_queryset(
                content_type=content_type,
                object_pk=object_pk_arg,
                site_id=site_id,
            )


class CommentCount(DefaultsMixin, generics.GenericAPIView):
    """Get number of comments posted to a given ContentType and object ID."""

    serializer_class = serializers.ReadCommentSerializer
    permission_classes = (permissions.AllowAny,)

    def get_queryset(self):
        content_type_arg = self.kwargs.get("content_type", None)
        object_pk_arg = self.kwargs.get("object_pk", None)
        app_label, model = content_type_arg.split("-")
        content_type = ContentType.objects.get_by_natural_key(app_label, model)
        site_id = getattr(settings, "SITE_ID", None)
        if not site_id:
            site_id = get_current_site_id(self.request)
        fkwds = {
            "content_type": content_type,
            "object_pk": object_pk_arg,
            "site__pk": site_id,
            "is_public": True,
        }
        if getattr(settings, "COMMENTS_HIDE_REMOVED", True):
            fkwds["is_removed"] = False
        return get_comment_model().objects.filter(**fkwds)

    def get(self, request, *args, **kwargs):
        return Response({"count": self.get_queryset().count()})


class CreateReportFlag(DefaultsMixin, generics.CreateAPIView):
    """Create 'removal suggestion' flags."""

    serializer_class = serializers.FlagSerializer
    permission_classes = (permissions.IsAuthenticatedOrReadOnly,)

    schema = AutoSchema(operation_id_base="ReportFlag")

    def post(self, request, *args, **kwargs):
        comment = get_object_or_404(
            get_comment_model(), pk=int(request.data["comment"])
        )
        check_option(comment, "comment_flagging_enabled")
        return super(CreateReportFlag, self).post(request, *args, **kwargs)

    def perform_create(self, serializer):
        perform_flag(self.request, serializer.validated_data["comment"])


class PostCommentReaction(mixins.CreateModelMixin, generics.GenericAPIView):
    """Create and delete comment reactions."""

    serializer_class = serializers.WriteCommentReactionSerializer
    permission_classes = (permissions.IsAuthenticated,)

    def post(self, request, *args, **kwargs):
        comment = get_object_or_404(
            get_comment_model(), pk=int(request.data["comment"])
        )
        check_option(comment, "comment_reactions_enabled")
        self.create(request, *args, **kwargs)
        # Create a new response object with the list of reactions the
        # comment has received. If other users sent reactions they all will
        # be reflected in the comment, not only the reaction sent with this
        # particular request.
        if self.created:
            return Response(
                comment.get_reactions(), status=status.HTTP_201_CREATED
            )
        else:
            return Response(comment.get_reactions(), status=status.HTTP_200_OK)

    def perform_create(self, serializer):
        cr_qs = CommentReaction.objects.filter(**serializer.validated_data)
        if cr_qs.filter(authors=self.request.user).count() == 1:
            self.created = False
            if cr_qs[0].counter == 1:
                cr_qs.delete()
            else:
                cr_qs.update(counter=F("counter") - 1)
                cr_qs[0].authors.remove(self.request.user)
        else:
            creaction, _ = CommentReaction.objects.get_or_create(
                **serializer.validated_data
            )
            creaction.authors.add(self.request.user)
            # self.created is True when the user reacting is added.
            self.created = True
            creaction.counter += 1
            creaction.save()



import logging
from collections import OrderedDict

from django.contrib.contenttypes.fields import GenericForeignKey
from django.contrib.contenttypes.models import ContentType
from django.contrib.sites.models import Site
from django.core import signing
from django.db import models
from django.db.models import F, Max, Min, Prefetch, Q
from django.db.models.signals import post_delete
from django.db.transaction import atomic
from django.urls import reverse
from django.utils.translation import gettext_lazy as _
from django_comments.managers import CommentManager
from django_comments.models import Comment, CommentFlag
from django_comments_ink import (
    caching,
    get_comment_reactions_enum,
    get_model,
    get_object_reactions_enum,
)
from django_comments_ink.conf import settings
from django_comments_ink.utils import get_current_site_id

logger = logging.getLogger(__name__)


def max_thread_level_for_content_type(content_type):
    app_model = "%s.%s" % (content_type.app_label, content_type.model)
    if app_model in settings.COMMENTS_INK_MAX_THREAD_LEVEL_BY_APP_MODEL:
        return settings.COMMENTS_INK_MAX_THREAD_LEVEL_BY_APP_MODEL[app_model]
    else:
        return settings.COMMENTS_INK_MAX_THREAD_LEVEL


class MaxThreadLevelExceededException(Exception):
    def __init__(self, comment):
        self.comment = comment
        # self.max_by_app = max_thread_level_for_content_type(content_type)

    def __str__(self):
        return "Max thread level reached for comment %d" % self.comment.id


class InkCommentManager(CommentManager):
    def for_app_models(self, *args, **kwargs):
        """Return comments for pairs "app.model" given in args"""
        content_types = []
        for app_model in args:
            app, model = app_model.split(".")
            content_types.append(
                ContentType.objects.get(app_label=app, model=model)
            )
        return self.for_content_types(content_types, **kwargs)

    def for_content_types(self, content_types, site=None):
        filter_fields = {"content_type__in": content_types}
        if site is not None:
            filter_fields["site"] = site
        qs = self.get_queryset().filter(**filter_fields).reverse()
        return qs

    def get_queryset(self):
        qs = super(InkCommentManager, self).get_queryset()
        return qs.select_related("user", "content_type").order_by(
            *settings.COMMENTS_INK_LIST_ORDER
        )


class InkComment(Comment):
    thread_id = models.IntegerField(default=0, db_index=True)
    parent_id = models.IntegerField(default=0)
    level = models.SmallIntegerField(default=0)
    order = models.IntegerField(default=1, db_index=True)
    followup = models.BooleanField(
        blank=True, default=False, help_text=_("Notify follow-up comments")
    )
    nested_count = models.IntegerField(default=0, db_index=True)
    objects = InkCommentManager()
    norel_objects = CommentManager()

    def get_absolute_url(self, anchor_pattern="#comment-%(id)s"):
        return reverse(
            "comments-url-redirect",
            args=(self.content_type_id, self.object_pk, self.pk),
        ) + (anchor_pattern % self.__dict__)

    def save(self, *args, **kwargs):
        caching.clear_cache(self.content_type.id, self.object_pk, self.site.pk)
        is_new = self.pk is None
        super(Comment, self).save(*args, **kwargs)
        if is_new:
            if not self.parent_id:
                self.parent_id = self.id
                self.thread_id = self.id
            else:
                if max_thread_level_for_content_type(self.content_type):
                    with atomic():
                        self._calculate_thread_data()
                else:
                    raise MaxThreadLevelExceededException(self)
            kwargs["force_insert"] = False
            super(Comment, self).save(*args, **kwargs)

    def _calculate_thread_data(self):
        # Implements the following approach:
        #  http://www.sqlteam.com/article/sql-for-threaded-discussion-forums
        parent = InkComment.objects.get(pk=self.parent_id)
        if parent.level == max_thread_level_for_content_type(self.content_type):
            raise MaxThreadLevelExceededException(self)

        self.thread_id = parent.thread_id
        self.level = parent.level + 1
        qc_eq_thread = InkComment.norel_objects.filter(
            thread_id=parent.thread_id
        )
        qc_ge_level = qc_eq_thread.filter(
            level__lte=parent.level, order__gt=parent.order
        )
        if qc_ge_level.count():
            min_order = qc_ge_level.aggregate(Min("order"))["order__min"]
            qc_eq_thread.filter(order__gte=min_order).update(
                order=F("order") + 1
            )
            self.order = min_order
        else:
            max_order = qc_eq_thread.aggregate(Max("order"))["order__max"]
            self.order = max_order + 1

        if self.id != parent.id:
            parent_ids = []
            while True:
                parent_ids.append(parent.pk)
                if parent.id == parent.parent_id:
                    break
                parent = qc_eq_thread.get(pk=parent.parent_id)
            if parent_ids:
                qc_eq_thread.filter(pk__in=parent_ids).update(
                    nested_count=F("nested_count") + 1
                )

    def get_reply_url(self):
        return reverse("comments-ink-reply", kwargs={"cid": self.pk})

    def allow_thread(self):
        if self.level < max_thread_level_for_content_type(self.content_type):
            return True
        else:
            return False

    def get_reactions(self):
        dci_cache = caching.get_cache()
        key = settings.COMMENTS_INK_CACHE_KEYS["comment_reactions"].format(
            comment_id=self.pk
        )
        if dci_cache != None and key != "":
            result = dci_cache.get(key)
            if result != None:
                logger.debug("Fetching %s from the cache", key)
                return result

        total_counter = 0
        reactions = OrderedDict([(k, {}) for k in get_comment_reactions_enum()])
        # First add the existing reactions sorted by reaction value.
        for item in self.reactions.order_by("reaction"):
            total_counter += item.counter
            reaction = get_comment_reactions_enum()(item.reaction)
            authors = [
                settings.COMMENTS_INK_API_USER_REPR(author)
                for author in item.authors.all()
            ]
            reactions[reaction.value] = {
                "value": reaction.value,
                "authors": authors,
                "counter": item.counter,
                "label": reaction.label,
                "icon": reaction.icon,
            }
        # Return only the values of OrderedDict after it's being sorted.
        result = {
            "counter": total_counter,
            "list": [v for k, v in reactions.items() if len(v)],
        }
        if dci_cache != None and key != "":
            dci_cache.set(key, result, timeout=None)
            logger.debug("Caching reactions for comment %d" % self.pk)
        return result

    @staticmethod
    def get_queryset(
        content_type=None, object_pk=None, content_object=None, site_id=None
    ):
        """
        Given either a content_object or the pair content_type and object_pk
        it returns the queryset with the comments posted to that object.
        """
        if content_object is None and (
            content_type is None or object_pk is None
        ):
            return None

        if content_object:
            content_type = ContentType.objects.get_for_model(content_object)
            object_pk = content_object.id

        flags = CommentFlag.objects.filter(
            flag__in=[CommentFlag.SUGGEST_REMOVAL]
        ).prefetch_related("user")

        reactions = CommentReaction.objects.all().prefetch_related("authors")

        prefetch_args = [
            Prefetch("flags", queryset=flags),
            Prefetch("reactions", queryset=reactions),
        ]

        fkwds = {
            "content_type": content_type,
            "object_pk": object_pk,
            "site__pk": site_id or get_current_site_id(),
            "is_public": True,
        }

        hide_removed = getattr(settings, "COMMENTS_HIDE_REMOVED", True)
        if hide_removed:
            fkwds["is_removed"] = False

        return (
            get_model().objects.prefetch_related(*prefetch_args).filter(**fkwds)
        )


def publish_or_withhold_nested_comments(comment, shall_be_public=False):
    qs = get_model().norel_objects.filter(
        ~Q(pk=comment.id), parent_id=comment.id
    )
    nested = [cm.id for cm in qs]
    qs.update(is_public=shall_be_public)
    while len(nested):
        cm_id = nested.pop()
        qs = get_model().norel_objects.filter(~Q(pk=cm_id), parent_id=cm_id)
        nested.extend([cm.id for cm in qs])
        qs.update(is_public=shall_be_public)
    # Update nested_count in parents comments in the same thread.
    # The comment.nested_count doesn't change because the comment's is_public
    # attribute is not changing, only its nested comments change, and it will
    # help to re-populate nested_count should it be published again.
    if shall_be_public:
        op = F("nested_count") + comment.nested_count
    else:
        op = F("nested_count") - comment.nested_count
    get_model().norel_objects.filter(
        thread_id=comment.thread_id,
        level__lt=comment.level,
        order__lt=comment.order,
    ).update(nested_count=op)


def publish_or_withhold_on_pre_save(sender, instance, raw, using, **kwargs):
    if not raw and instance and instance.id:
        shall_be_public = (not instance.is_removed) and instance.is_public
        publish_or_withhold_nested_comments(instance, shall_be_public)


def on_comment_deleted(sender, instance, using, **kwargs):
    # Create the list of nested ink-comments that have to deleted too.
    qs = get_model().norel_objects.filter(
        ~Q(pk=instance.id), parent_id=instance.id
    )
    nested = [cm.id for cm in qs]
    for cm_id in nested:
        qs = get_model().norel_objects.filter(~Q(pk=cm_id), parent_id=cm_id)
        nested.extend([cm.id for cm in qs])

    # Update the nested_count attribute up the tree.
    get_model().norel_objects.filter(
        thread_id=instance.thread_id,
        level__lt=instance.level,
        order__lt=instance.order,
    ).update(nested_count=F("nested_count") - instance.nested_count - 1)

    # Delete all reactions, and reaction authors, associated
    # with nested instances.
    creactions_qs = CommentReaction.objects.filter(comment__pk__in=nested)
    creactions_ids = [cr.id for cr in creactions_qs]

    CommentReactionAuthor.objects.filter(
        reaction__pk__in=creactions_ids
    )._raw_delete(using)
    creactions_qs._raw_delete(using)

    # Delete all the comments down the tree from instance.
    get_model().objects.filter(pk__in=nested)._raw_delete(using)


post_delete.connect(on_comment_deleted, sender=InkComment)

# ----------------------------------------------------------------------


class DummyDefaultManager:
    """
    Dummy Manager to mock django's CommentForm.check_for_duplicate method.
    """

    def __getattr__(self, name):
        return lambda *args, **kwargs: []

    def using(self, *args, **kwargs):
        return self


class TmpInkComment(dict):
    """
    Temporary InkComment to be pickled, ziped and appended to a URL.
    """

    _default_manager = DummyDefaultManager()

    def __getattr__(self, key):
        try:
            return self[key]
        except KeyError:
            return None

    def __setattr__(self, key, value):
        self[key] = value

    def save(self, *args, **kwargs):
        pass

    def _get_pk_val(self):
        if self.ink_comment:
            return self.ink_comment._get_pk_val()
        else:
            content_type = "%s.%s" % self.content_type.natural_key()
            return signing.dumps("%s:%s" % (content_type, self.object_pk))

    def __setstate__(self, state):
        ct_key = state.pop("content_type_key")
        ctype = ContentType.objects.get_by_natural_key(*ct_key)
        self.update(
            state,
            content_type=ctype,
            content_object=ctype.get_object_for_this_type(
                pk=state["object_pk"]
            ),
        )

    def __reduce__(self):
        state = {k: v for k, v in self.items() if k != "content_object"}
        ct = state.pop("content_type")
        state["content_type_key"] = ct.natural_key()
        return (TmpInkComment, (), state)


# ----------------------------------------------------------------------
class BlackListedDomain(models.Model):
    """
    A blacklisted domain from which comments should be discarded.
    Automatically populated with a small amount of spamming domains,
    gathered from http://www.joewein.net/spam/blacklist.htm

    You can download for free a recent version of the list, and subscribe
    to get notified on changes. Changes can be fetched with rsync for a
    small fee (check their conditions, or use any other Spam filter).
    """

    domain = models.CharField(max_length=200, db_index=True)

    def __str__(self):
        return self.domain

    class Meta:
        ordering = ("domain",)


# ----------------------------------------------------------------------
class BaseReactionEnum(models.TextChoices):
    @classmethod
    def set_icons(cls, icons):
        cls._icons = icons

    @property
    def icon(self):
        return self._icons[self]


class ReactionEnum(BaseReactionEnum):
    LIKE_IT = "+", "+1"
    DISLIKE_IT = "-", "-1"


ReactionEnum.set_icons(
    {ReactionEnum.LIKE_IT: "#128077", ReactionEnum.DISLIKE_IT: "#128078"}
)


class ReactionField(models.TextField):
    description = "Code representing a user reaction to a comment"

    def deconstruct(self):
        name, path, args, kwargs = super(ReactionField, self).deconstruct()
        kwargs.pop("choices", None)  # Ignore choice changes in migrations.
        return name, path, args, kwargs


# It allows the implementation of two models:
#  * CommentReaction
#  * ObjectReaction

# -----------------------------------------------
# Comment reaction model.


class CommentReaction(models.Model):
    reaction = ReactionField(
        _("reaction"),
        db_index=True,
        choices=get_comment_reactions_enum().choices,
    )
    comment = models.ForeignKey(
        get_model(),
        verbose_name=_("reactions"),
        related_name="reactions",
        on_delete=models.CASCADE,
    )
    counter = models.IntegerField(default=0)
    authors = models.ManyToManyField(
        settings.AUTH_USER_MODEL,
        through="CommentReactionAuthor",
        through_fields=("reaction", "author"),
    )

    class Meta:
        verbose_name = _("comment reactions")
        verbose_name_plural = _("comments reactions")

    def delete_from_cache(self):
        dci_cache = caching.get_cache()
        key = settings.COMMENTS_INK_CACHE_KEYS["comment_reactions"].format(
            comment_id=self.comment.pk
        )
        if dci_cache != None and key != "" and dci_cache.get(key):
            logger.debug(
                "Delete cached list of comment reactions in key %s" % key
            )
            dci_cache.delete(key)

    def save(self, *args, **kwargs):
        self.delete_from_cache()
        super(CommentReaction, self).save(*args, **kwargs)


class CommentReactionAuthor(models.Model):
    reaction = models.ForeignKey(CommentReaction, on_delete=models.CASCADE)
    author = models.ForeignKey(
        settings.AUTH_USER_MODEL, on_delete=models.CASCADE
    )


# -----------------------------------------------
# Object reaction model.


class ObjectReaction(models.Model):
    reaction = ReactionField(
        _("reaction"),
        db_index=True,
        choices=get_object_reactions_enum().choices,
    )
    content_type = models.ForeignKey(
        ContentType,
        verbose_name=_("content type"),
        related_name="content_type_set_for_%(class)s",
        on_delete=models.CASCADE,
    )
    object_pk = models.CharField(_("object ID"), db_index=True, max_length=64)
    content_object = GenericForeignKey(
        ct_field="content_type", fk_field="object_pk"
    )
    site = models.ForeignKey(Site, on_delete=models.CASCADE)
    counter = models.IntegerField(default=0)
    authors = models.ManyToManyField(
        settings.AUTH_USER_MODEL,
        through="ObjectReactionAuthor",
        through_fields=("reaction", "author"),
    )

    class Meta:
        verbose_name = _("object reactions")
        verbose_name_plural = _("objects reactions")

    def delete_from_cache(self):
        dci_cache = caching.get_cache()
        key = settings.COMMENTS_INK_CACHE_KEYS["object_reactions"].format(
            ctype_pk=self.content_type.id,
            object_pk=self.object_pk,
            site_id=self.site.id,
        )
        if dci_cache != None and key != "" and dci_cache.get(key):
            logger.debug(
                "Delete cached list of object reactions in key %s" % key
            )
            dci_cache.delete(key)

    def save(self, *args, **kwargs):
        self.delete_from_cache()
        super(ObjectReaction, self).save(*args, **kwargs)


class ObjectReactionAuthor(models.Model):
    reaction = models.ForeignKey(ObjectReaction, on_delete=models.CASCADE)
    author = models.ForeignKey(
        settings.AUTH_USER_MODEL, on_delete=models.CASCADE
    )



from django_comments_ink.api.views import (
    CommentCount,
    CommentCreate,
    CommentList,
    CreateReportFlag,
    PostCommentReaction,
)

__all__ = (
    CommentCreate,
    CommentList,
    CommentCount,
    PostCommentReaction,
    CreateReportFlag,
)



import logging

from django.core.cache import InvalidCacheBackendError, caches
from django_comments_ink.conf import settings

logger = logging.getLogger(__name__)

dci_cache = None


def get_cache():
    global dci_cache
    if dci_cache != None:
        return dci_cache

    try:
        dci_cache = caches[settings.COMMENTS_INK_CACHE_NAME]
    except InvalidCacheBackendError:
        logger.warning(
            "Cache '%s' is not defined in the settings module, "
            "using 'default' instead.",
            settings.COMMENTS_INK_CACHE_NAME,
        )
        dci_cache = caches["default"]

    return dci_cache


def clear_cache(content_type_id, object_pk, site_id):
    dci_cache = get_cache()
    if dci_cache == None:
        logger.warning(
            "Cannot access the cache. Could not clear the cache "
            "for content_type={%d}, object_pk={%d} and site_id={%d}. "
            % (content_type_id, object_pk, site_id)
        )
        return False

    keys = []
    for key_pattern in [
        "comments_qs",
        "comments_count",
        "comments_paged",
    ]:
        key = settings.COMMENTS_INK_CACHE_KEYS[key_pattern].format(
            ctype_pk=content_type_id, object_pk=object_pk, site_id=site_id
        )
        keys.append(key)

    for key in keys:
        if dci_cache.get(key):
            logger.debug("Delete cached key %s" % key)
            dci_cache.delete(key)

    return True



from __future__ import unicode_literals

import json
from datetime import datetime
from unittest.mock import Mock, patch

import django_comments
import pytest
import pytz
from django.contrib.auth.models import AnonymousUser, User
from django.contrib.contenttypes.models import ContentType
from django.contrib.sites.models import Site
from django.test import TestCase
from django.urls import reverse
from django_comments.moderation import CommentModerator
from django_comments.signals import comment_will_be_posted
from django_comments_ink.api.serializers import (
    FlagSerializer,
    ReadCommentSerializer,
    WriteCommentReactionSerializer,
    WriteCommentSerializer,
)
from django_comments_ink.conf import settings
from django_comments_ink.models import InkComment
from django_comments_ink.moderation import moderator
from django_comments_ink.signals import should_request_be_authorized
from django_comments_ink.tests.models import (
    Article,
    authorize_api_post_comment,
    comment_will_be_rejected,
)
from django_comments_ink.tests.utils import post_comment
from rest_framework.test import APIClient


class FakeRequest:
    def __init__(self, user):
        self.user = user
        self.auth = None


class WriteCommentSerializerAsVisitorTestCase(TestCase):
    def setUp(self):
        self.patcher = patch("django_comments_ink.views.utils.send_mail")
        self.mock_mailer = self.patcher.start()
        self.article = Article.objects.create(
            title="October", slug="october", body="What I did on October..."
        )
        self.form = django_comments.get_form()(self.article)
        # Remove the following fields on purpose, as we don't know them and
        # therefore we don't send them when using the web API (unless when)
        # using the JavaScript plugin, but that is not the case being tested
        # here.
        for field_name in ["security_hash", "timestamp"]:
            self.form.initial.pop(field_name)

    def tearDown(self):
        self.patcher.stop()

    def test_post_comment_before_connecting_signal(self):
        data = {
            "name": "Joe Bloggs",
            "email": "joe@bloggs.com",
            "followup": True,
            "reply_to": 0,
            "comment": "This post comment request should fail",
        }
        data.update(self.form.initial)
        response = post_comment(data)
        self.assertEqual(response.status_code, 403)
        self.assertEqual(
            response.rendered_content,
            b'{"detail":"You do not have permission to perform this action."}',
        )
        self.assertTrue(self.mock_mailer.call_count == 0)

    def test_post_comment_after_connecting_signal(self):
        should_request_be_authorized.connect(authorize_api_post_comment)
        data = {
            "name": "Joe Bloggs",
            "email": "joe@bloggs.com",
            "followup": True,
            "reply_to": 0,
            "comment": "This post comment request should fail",
        }
        data.update(self.form.initial)
        client = APIClient()
        token = "Token 08d9fd42468aebbb8087b604b526ff0821ce4525"
        client.credentials(HTTP_AUTHORIZATION=token)
        self.assertTrue(self.mock_mailer.call_count == 0)
        response = client.post(reverse("comments-ink-api-create"), data)
        self.assertEqual(response.status_code, 204)  # Confirmation req sent.
        self.assertTrue(self.mock_mailer.call_count == 1)
        should_request_be_authorized.disconnect(authorize_api_post_comment)


class WriteCommentSerializerTestCase(TestCase):
    def setUp(self):
        self.patcher = patch("django_comments_ink.views.utils.send_mail")
        self.mock_mailer = self.patcher.start()
        self.user = User.objects.create_user(
            "joe", "joe@bloggs.com", "pwd", first_name="Joe", last_name="Bloggs"
        )
        self.article = Article.objects.create(
            title="October", slug="october", body="What I did on October..."
        )
        self.form = django_comments.get_form()(self.article)
        # Remove the following fields on purpose, as we don't know them and
        # therefore we don't send them when using the web API (unless when)
        # using the JavaScript plugin, but that is not the case being tested
        # here.
        for field_name in ["security_hash", "timestamp"]:
            self.form.initial.pop(field_name)
        should_request_be_authorized.connect(authorize_api_post_comment)

    def tearDown(self):
        self.patcher.stop()
        should_request_be_authorized.disconnect(authorize_api_post_comment)

    def test_post_comment_as_registered_user_after_connecting_signal(self):
        data = {
            "name": "",
            "email": "",
            "followup": True,
            "reply_to": 0,
            "comment": "This post comment request should fail",
        }
        data.update(self.form.initial)
        client = APIClient()
        client.login(username="joe", password="pwd")
        self.assertTrue(self.mock_mailer.call_count == 0)
        response = client.post(reverse("comments-ink-api-create"), data)
        self.assertEqual(response.status_code, 201)  # Comment created.
        self.assertTrue(self.mock_mailer.call_count == 0)

    def test_post_comment_can_be_rejected(self):
        comment_will_be_posted.connect(comment_will_be_rejected)
        data = {
            "name": "Joe Bloggs",
            "email": "joe@bloggs.com",
            "followup": True,
            "reply_to": 0,
            "comment": "This post comment request should fail",
        }
        data.update(self.form.initial)
        client = APIClient()
        client.login(username="joe", password="pwd")
        self.assertTrue(self.mock_mailer.call_count == 0)
        response = client.post(reverse("comments-ink-api-create"), data)
        self.assertEqual(response.status_code, 403)  # Comment created.
        self.assertTrue(self.mock_mailer.call_count == 0)
        comment_will_be_posted.disconnect(comment_will_be_rejected)

    def test_post_comment_can_be_put_in_moderation(self):  # code 202.
        # Set article's date back to 1990. And set up moderation for old
        # objects.
        self.article.publish = datetime(1990, 1, 1)
        self.article.save()

        # Create the moderator class, and register it so that comments
        # posted to Article are moderated using the class
        # PostCommentModerator.
        class PostCommentModerator(CommentModerator):
            email_notification = False
            auto_moderate_field = "publish"
            moderate_after = 365

        moderator.register(Article, PostCommentModerator)

        data = {
            "name": "Joe Bloggs",
            "email": "joe@bloggs.com",
            "followup": True,
            "reply_to": 0,
            "comment": "This post comment request should fail",
        }
        data.update(self.form.initial)
        client = APIClient()
        client.login(username="joe", password="pwd")
        self.assertTrue(self.mock_mailer.call_count == 0)
        response = client.post(reverse("comments-ink-api-create"), data)
        self.assertEqual(response.status_code, 202)  # Comment created.
        self.assertTrue(self.mock_mailer.call_count == 0)

    def test_validate_name_without_value(self):
        data = {
            "name": "",
            "email": "joe@bloggs.com",
            "followup": True,
            "reply_to": 0,
            "comment": "This is a comment",
        }
        data.update(self.form.initial)
        mock_user = Mock(spec=self.user, autospec=True, return_value=self.user)
        mock_user.configure_mock(
            **{"get_username.return_value": self.user.username}
        )
        delattr(mock_user, "get_full_name")
        req = FakeRequest(mock_user)
        ser = WriteCommentSerializer(data=data, context={"request": req})
        self.assertTrue(ser.is_valid())
        self.assertEqual(ser.data["name"], self.user.username)

    def test_validate_reply_to_does_not_exist(self):
        data = {
            "name": "Joe Bloggs",
            "email": "joe@bloggs.com",
            "followup": True,
            "reply_to": 1,
            "comment": "This is a comment",
        }
        data.update(self.form.initial)
        req = FakeRequest(AnonymousUser())
        ser = WriteCommentSerializer(data=data, context={"request": req})
        self.assertFalse(ser.is_valid())
        expected_errors = '{"reply_to": ["reply_to comment does not exist"]}'
        self.assertEqual(json.dumps(ser.errors), expected_errors)

    def test_content_type_or_object_pk_cant_be_None(self):
        data = {
            "name": "Joe Bloggs",
            "email": "joe@bloggs.com",
            "followup": True,
            "reply_to": 0,
            "comment": "This post comment request should fail",
        }
        # Make content_type None.
        data.update(self.form.initial)
        data["content_type"] = None
        req = FakeRequest(AnonymousUser())
        ser = WriteCommentSerializer(data=data, context={"request": req})
        self.assertFalse(ser.is_valid())
        expected_errors = '{"content_type": ["This field may not be null."]}'
        self.assertEqual(json.dumps(ser.errors), expected_errors)

        # Make object_pk None.
        data.update(self.form.initial)
        data["object_pk"] = None
        ser = WriteCommentSerializer(data=data, context={"request": None})
        self.assertFalse(ser.is_valid())
        expected_errors = '{"object_pk": ["This field may not be null."]}'
        self.assertEqual(json.dumps(ser.errors), expected_errors)

    def test_validate_forging_content_type_raises_LookupError(self):
        data = {
            "name": "Joe Bloggs",
            "email": "joe@bloggs.com",
            "followup": True,
            "reply_to": 0,
            "comment": "This post comment request should fail",
        }
        data.update(self.form.initial)
        data["content_type"] = "doesnot.exist"
        req = FakeRequest(AnonymousUser())
        ser = WriteCommentSerializer(data=data, context={"request": req})
        self.assertFalse(ser.is_valid())
        expected_errors = (
            '{"non_field_errors": ["Invalid content_type '
            "value: 'doesnot.exist'\"]}"
        )
        self.assertEqual(json.dumps(ser.errors), expected_errors)

    def test_validate_forging_content_type_raises_model_DoesNotExist(self):
        data = {
            "name": "Joe Bloggs",
            "email": "joe@bloggs.com",
            "followup": True,
            "reply_to": 0,
            "comment": "This post comment request should fail",
        }
        data.update(self.form.initial)
        data["content_type"] = "auth.user"
        data["object_pk"] = "2"
        req = FakeRequest(AnonymousUser())
        ser = WriteCommentSerializer(data=data, context={"request": req})
        self.assertFalse(ser.is_valid())
        expected_errors = (
            '{"non_field_errors": ["No object matching '
            "content-type 'auth.user' and object PK '2'.\"]}"
        )
        self.assertEqual(json.dumps(ser.errors), expected_errors)

    def test_validate_forging_object_pk_raises_ValueError(self):
        data = {
            "name": "Joe Bloggs",
            "email": "joe@bloggs.com",
            "followup": True,
            "reply_to": 0,
            "comment": "This post comment request should fail",
        }
        data.update(self.form.initial)
        data["object_pk"] = "tal"
        req = FakeRequest(AnonymousUser())
        ser = WriteCommentSerializer(data=data, context={"request": req})
        self.assertFalse(ser.is_valid())
        expected_errors = (
            '{"non_field_errors": ["Attempting to get '
            "content-type 'tests.article' and object "
            "PK 'tal' raised ValueError\"]}"
        )
        self.assertEqual(json.dumps(ser.errors), expected_errors)


def get_fake_avatar(comment):
    return f"/fake/avatar/{comment.user.username}"


funcpath = "django_comments_ink.tests.test_serializers.get_fake_avatar"


class ReadCommentsGetUserAvatarTestCase(TestCase):
    # Test ReadCommentSerializer method get_user_avatar.
    # Change setting comments_ink_API_GET_USER_AVATAR so that it uses a
    # deterministic function: get_fake_avatar (here defined). Then send a
    # couple of comments and verify that the function is called.

    def setUp(self):
        joe = User.objects.create_user(
            "joe",
            "joe@example.com",
            "joepwd",
            first_name="Joe",
            last_name="Bloggs",
        )
        alice = User.objects.create_user(
            "alice",
            "alice@tal.com",
            "alicepwd",
            first_name="Alice",
            last_name="Bloggs",
        )
        self.article = Article.objects.create(
            title="September", slug="september", body="During September..."
        )
        self.article_ct = ContentType.objects.get(
            app_label="tests", model="article"
        )
        self.site = Site.objects.get(pk=1)

        # Testing comment from Bob.
        InkComment.objects.create(
            content_type=self.article_ct,
            object_pk=self.article.id,
            content_object=self.article,
            site=self.site,
            comment="testing comment from Bob",
            user=joe,
            submit_date=datetime.now(),
        )

        # Testing comment from Alice.
        InkComment.objects.create(
            content_type=self.article_ct,
            object_pk=self.article.id,
            content_object=self.article,
            site=self.site,
            comment="testing comment from Alice",
            user=alice,
            submit_date=datetime.now(),
        )


class RenderSubmitDateTestCase(TestCase):
    def setUp(self):
        self.article = Article.objects.create(
            title="October", slug="october", body="What I did on October..."
        )

    def create_comment(self, submit_date_is_aware=True):
        site = Site.objects.get(pk=1)
        ctype = ContentType.objects.get(app_label="tests", model="article")
        if submit_date_is_aware:
            utc = pytz.timezone("UTC")
            submit_date = datetime(2021, 1, 10, 10, 15, tzinfo=utc)
        else:
            submit_date = datetime(2021, 1, 10, 10, 15)
        self.cm = InkComment.objects.create(
            content_type=ctype,
            object_pk=self.article.id,
            content_object=self.article,
            site=site,
            name="Joe Bloggs",
            email="joe@bloggs.com",
            comment="Just a comment",
            submit_date=submit_date,
        )

    @patch.multiple("django.conf.settings", USE_TZ=False)
    @patch.multiple("django_comments_ink.conf.settings", USE_TZ=False)
    def test_submit_date_when_use_tz_is_false(self):
        self.create_comment(submit_date_is_aware=False)
        qs = InkComment.objects.all()
        ser = ReadCommentSerializer(qs, context={"request": None}, many=True)
        self.assertEqual(
            ser.data[0]["submit_date"], "Jan. 10, 2021, 10:15 a.m."
        )

    @patch.multiple("django.conf.settings", USE_TZ=True)
    @patch.multiple("django_comments_ink.conf.settings", USE_TZ=True)
    def test_submit_date_when_use_tz_is_true(self):
        self.create_comment(submit_date_is_aware=True)
        qs = InkComment.objects.all()
        ser = ReadCommentSerializer(qs, context={"request": None}, many=True)
        self.assertEqual(
            ser.data[0]["submit_date"], "Jan. 10, 2021, 11:15 a.m."
        )


# ---------------------------------------------------------------------
# Tests for FlagSerializer. Using pytest instead of unittest.


@pytest.mark.django_db
def test_flag_serializer_is_valid(an_articles_comment):
    data = {"comment": an_articles_comment.pk, "flag": "report"}
    ser = FlagSerializer(data=data)
    assert ser.is_valid()


@pytest.mark.django_db
def test_flag_serializer_is_not_valid(an_articles_comment):
    data = {"comment": an_articles_comment.pk, "flag": "non-supported-flag"}
    ser = FlagSerializer(data=data)
    assert not ser.is_valid()


@pytest.mark.django_db
def test_ReadReactionsField(a_comments_reaction, an_user):
    context = {"request": None}
    ser = ReadCommentSerializer(a_comments_reaction.comment, context=context)
    assert len(ser.data["reactions"]) == 1
    reaction = ser.data["reactions"][0]
    assert "reaction" in reaction and reaction["reaction"] == "+"
    assert "label" in reaction and reaction["label"] == "+1"
    assert "icon" in reaction and reaction["icon"] == "#128077"
    assert "authors" in reaction and len(reaction["authors"]) == 1
    author = reaction["authors"][0]
    assert author["id"] == an_user.id
    assert author["author"] == settings.COMMENTS_INK_API_USER_REPR(an_user)


@pytest.mark.django_db
def test_ReadCommentSerializer_get_flags(a_comments_flag):
    context = {"request": None}
    ser = ReadCommentSerializer(a_comments_flag.comment, context=context)
    assert len(ser.data["flags"]) == 1
    flag = ser.data["flags"][0]
    assert flag["flag"] == "removal"
    assert flag["id"] == a_comments_flag.user.id
    user_repr = settings.COMMENTS_INK_API_USER_REPR(a_comments_flag.user)
    assert flag["user"] == user_repr


@pytest.mark.django_db
def test_WriteCommentReactionSerializer(an_articles_comment):
    # 1st: Test a non-existing reaction is caught by the serializer.
    reaction = "?"
    ser = WriteCommentReactionSerializer(
        data={"reaction": reaction, "comment": an_articles_comment}
    )
    assert ser.is_valid() == False
    assert "reaction" in ser.errors
    assert ser.errors["reaction"][0].code == "invalid_choice"

    # 2nd: Test an existing reaction makes the serializer valid.
    reaction = "+"
    ser = WriteCommentReactionSerializer(
        data={"reaction": reaction, "comment": an_articles_comment}
    )
    assert ser.is_valid() == True
